#include <stdlib.h>
#include <string.h>
#include <curses.h> /* a commenter pour compiler pour mac/linux */
/*#include <ncurses.h> */ /* a decommenter pour compiler pour mac/linux */
#define ERR_FICH_INVALIDE -1
#define ERR_FICH_INTROUVABLE -2
#define ERR_MEMOIRE -3
#define SUCCES 0

#define HAUTEUR_FEN 26
#define LARGEUR_FEN 60
#define HAUTEUR_MAX 20
#define LARGEUR_MAX 30
#define NIV_LONG_MAX_NOM 25

#define POS_TITRE_X (LARGEUR_FEN / 2 - 8)
#define POS_TITRE_Y 1
#define POS_NOM_X 1
#define POS_NOM_Y 5
#define POS_INFOS_X (LARGEUR_MAX + POS_JEU_X + 2)
#define POS_INFOS_Y 7
#define POS_INSTRU_X (LARGEUR_MAX + POS_JEU_X + 2)
#define POS_INSTRU_Y 15
#define POS_JEU_X 1
#define POS_JEU_Y 6
#define TITRE_COULEUR 8

#define VIDE 0   /* Ces valeurs ne sont pas du tout mises au bol !  */
#define MUR 1    /* Elles sont definies de tel sorte que la gestion */
#define JOUEUR 2 /* des collisions est simplifiee.                  */
#define BLOC 3
#define TROU 4 
#define JOUEUR_TROU 6
#define BLOC_TROU 7
#define VIDE_COULEUR 5 /* La couleur 0 ne marche pas */

#define C_VIDE ' '
#define C_MUR ''
#define C_JOUEUR 'J'          
#define C_BLOC '#'
#define C_TROU '*'

#define NO_EVEN 0
#define HAUT 1
#define GAUCHE 2
#define DROITE 3
#define BAS 4
#define QUITTER 5
#define RECOMMENCER 6

typedef struct
{
    short x;
    short y;
}Pos;

typedef struct
{
    Pos posJ;    
    short nbTrous, /* utilise seulement lors du chargement */
          nbCaisses,
          nbCaisseTrou,
          nbDepl,
          largMat,
          hautMat;
    char **mat;
    char nom[NIV_LONG_MAX_NOM];
}Niveau;

int nivCharger(char *nomFichier, int numNiv, Niveau *nivAct);
int nivChargerCorps(FILE *fNiv, Niveau *nivAct);
int nivCopier(Niveau *nSrc, Niveau *nDest);
void nivLiberer(Niveau *niv);
void bouger(Niveau *nivAct, int direction);
void afficherTout(Niveau nivAct);
void afficherCadre(Niveau nivAct, Pos debutCadre, Pos finCadre);
void afficherTitre(void);
void afficherInfos(Niveau nivAct);
int recupEvent(void);
int init(Niveau *nivAct);
void message(char *texte, int nbSec);

int main(void)
{
    int action = NO_EVEN,
        niveau = 1,
        retour;
    Pos debCadre = {0, 0},
        finCadre = {0, 0};
    Niveau nivAct,
           savNivAct;
    init(&nivAct);
    goto DEBUT;
    
    while(action != QUITTER)
    {
        if(action == RECOMMENCER)
        {
            nivCopier(&savNivAct, &nivAct);
            afficherTout(nivAct);
        }
        else
        {
            bouger(&nivAct, action);
            afficherCadre(nivAct, debCadre, finCadre);
            afficherInfos(nivAct);
        }
        if(nivAct.nbCaisses == 0)
        {
            message("BRAVO! NIVEAU REUSSI!", 3);
            niveau += 1;
            nivLiberer(&nivAct);
            nivLiberer(&savNivAct);
            DEBUT:
            retour = nivCharger("niveaux.skb", niveau, &nivAct);
            if(retour == ERR_MEMOIRE) message("Erreur d'allocation.", 3);
            else if(retour == ERR_FICH_INTROUVABLE) message("Fichier introuvable.", 3);
            else if(retour == ERR_FICH_INVALIDE) message("Fichier invalide.", 3);
            if(nivCopier(&nivAct, &savNivAct) == ERR_MEMOIRE)
                message("Erreur d'allocation.", 3);                  
            finCadre.x = nivAct.largMat;
            finCadre.y = nivAct.hautMat;
            afficherTout(nivAct);
        }
        action = recupEvent();    
    }
    message("A la prochaine !", 2);
    
    nivLiberer(&nivAct);
    nivLiberer(&savNivAct);
    return 0;
}

/**********************************************/
int init(Niveau *nivAct)
{
    initscr();
    curs_set(0);
    noecho();
    resize_term(HAUTEUR_FEN, LARGEUR_FEN);
    start_color();
    init_pair(VIDE_COULEUR, COLOR_BLACK, COLOR_BLACK);
    init_pair(BLOC, COLOR_YELLOW, COLOR_BLACK);
    init_pair(TROU, COLOR_GREEN, COLOR_BLACK);
    init_pair(BLOC_TROU, COLOR_YELLOW, COLOR_BLACK);
    init_pair(MUR, COLOR_RED, COLOR_BLACK);
    init_pair(JOUEUR, COLOR_CYAN, COLOR_BLACK);
    init_pair(JOUEUR_TROU, COLOR_BLUE, COLOR_BLACK);
    init_pair(TITRE_COULEUR, COLOR_GREEN, COLOR_BLUE);
    keypad(stdscr, TRUE);
    
    return 0;
}

/**********************************************/
void bouger(Niveau *nivAct, int direction)
{
    nivAct->mat[nivAct->posJ.y][nivAct->posJ.x] -= JOUEUR; /* On remet le joueur a la fin */
    nivAct->nbDepl += 1;
    switch(direction)
    {
        case HAUT:
            if(nivAct->mat[nivAct->posJ.y - 1][nivAct->posJ.x] % 2 == VIDE)
                nivAct->posJ.y -= 1;
            else if(nivAct->mat[nivAct->posJ.y - 1][nivAct->posJ.x] % 4 == BLOC &&
                    nivAct->mat[nivAct->posJ.y - 2][nivAct->posJ.x] % 2 == VIDE)
            {
                nivAct->posJ.y -= 1;
                nivAct->mat[nivAct->posJ.y][nivAct->posJ.x] -= BLOC;
                nivAct->mat[nivAct->posJ.y - 1][nivAct->posJ.x] += BLOC;
                if(nivAct->mat[nivAct->posJ.y][nivAct->posJ.x] == TROU)
                {
                    nivAct->nbCaisseTrou -= 1;
                    nivAct->nbCaisses += 1;
                }
                if(nivAct->mat[nivAct->posJ.y - 1][nivAct->posJ.x] == BLOC_TROU)
                {
                    nivAct->nbCaisseTrou += 1;
                    nivAct->nbCaisses -= 1;
                }
            }
            else /* Si le joueur etait bloque (mur, caisse-caisse, caisse-mur */
                nivAct->nbDepl -= 1;
            break;
        case GAUCHE:
            if(nivAct->mat[nivAct->posJ.y][nivAct->posJ.x - 1] % 2 == VIDE)
                nivAct->posJ.x -= 1;
            else if(nivAct->mat[nivAct->posJ.y][nivAct->posJ.x - 1] % 4 == BLOC &&
                    nivAct->mat[nivAct->posJ.y][nivAct->posJ.x - 2] % 2 == VIDE)
            {
                nivAct->posJ.x -= 1;
                nivAct->mat[nivAct->posJ.y][nivAct->posJ.x] -= BLOC;
                nivAct->mat[nivAct->posJ.y][nivAct->posJ.x - 1] += BLOC;
                if(nivAct->mat[nivAct->posJ.y][nivAct->posJ.x] == TROU)
                {
                    nivAct->nbCaisseTrou -= 1;
                    nivAct->nbCaisses += 1;
                }
                if(nivAct->mat[nivAct->posJ.y][nivAct->posJ.x - 1] == BLOC_TROU)
                {
                    nivAct->nbCaisseTrou += 1;
                    nivAct->nbCaisses -= 1;
                }
            }
            else
                nivAct->nbDepl -= 1;
            break;
        case DROITE:
            if(nivAct->mat[nivAct->posJ.y][nivAct->posJ.x + 1] % 2 == VIDE)
                nivAct->posJ.x += 1;
            else if(nivAct->mat[nivAct->posJ.y][nivAct->posJ.x + 1] % 4 == BLOC &&
                    nivAct->mat[nivAct->posJ.y][nivAct->posJ.x + 2] % 2 == VIDE)
            {
                nivAct->posJ.x += 1;
                nivAct->mat[nivAct->posJ.y][nivAct->posJ.x] -= BLOC;
                nivAct->mat[nivAct->posJ.y][nivAct->posJ.x + 1] += BLOC;
                if(nivAct->mat[nivAct->posJ.y][nivAct->posJ.x] == TROU)
                {
                    nivAct->nbCaisseTrou -= 1;
                    nivAct->nbCaisses += 1;
                }
                if(nivAct->mat[nivAct->posJ.y][nivAct->posJ.x + 1] == BLOC_TROU)
                {
                    nivAct->nbCaisseTrou += 1;
                    nivAct->nbCaisses -= 1;
                }
            }
            else
                nivAct->nbDepl -= 1;
            break;
        case BAS:
            if(nivAct->mat[nivAct->posJ.y + 1][nivAct->posJ.x] % 2 == VIDE)
                nivAct->posJ.y += 1;
            else if(nivAct->mat[nivAct->posJ.y + 1][nivAct->posJ.x] % 4 == BLOC &&
                    nivAct->mat[nivAct->posJ.y + 2][nivAct->posJ.x] % 2 == VIDE)
            {
                nivAct->posJ.y += 1;
                nivAct->mat[nivAct->posJ.y][nivAct->posJ.x] -= BLOC;
                nivAct->mat[nivAct->posJ.y + 1][nivAct->posJ.x] += BLOC;
                if(nivAct->mat[nivAct->posJ.y][nivAct->posJ.x] == TROU)
                {
                    nivAct->nbCaisseTrou -= 1;
                    nivAct->nbCaisses += 1;
                }
                if(nivAct->mat[nivAct->posJ.y + 1][nivAct->posJ.x] == BLOC_TROU)
                {
                    nivAct->nbCaisseTrou += 1;
                    nivAct->nbCaisses -= 1;
                }
            }
            else
                nivAct->nbDepl -= 1;
            break;
        default:
            nivAct->nbDepl -= 1;
            break;
    }
    nivAct->mat[nivAct->posJ.y][nivAct->posJ.x] += JOUEUR;
}

/**********************************************/
int nivCharger(char *nomFichier, int numNiv, Niveau *nivAct)
{
    FILE *fichNiv = NULL;
    char iFich = 0;
    int iNiv = 1,
        iNom = 0,
        retour;

    fichNiv = fopen(nomFichier, "r");
        if(fichNiv == NULL) return ERR_FICH_INTROUVABLE;
    
    while(iNiv < numNiv) /* On avance jusqu'au niveau voulu */
    {
        while(fgetc(fichNiv) != '\n'); /* Lecture du titre */
        /* Lecture du corps */
        while((iFich = fgetc(fichNiv)) == ' ' || iFich == '#' || iFich == '\n' ||
                  iFich == '*' || iFich == '$' || iFich == '.' || iFich == '@' || iFich == '+');
        iNiv += 1;
    }
    if(numNiv > 1)
        nivAct->nom[iNom++] = iFich;
    /* Enregistrement du titre du niveau voulu */
    while((iFich = fgetc(fichNiv)) != '\n' && iNom < NIV_LONG_MAX_NOM - 1)
        nivAct->nom[iNom++] = iFich;
    nivAct->nom[iNom++] = 0;
    if(iFich != '\n')
        while(fgetc(fichNiv) != '\n');
    
    /* Enregistrement des donnees du niveau */
    retour = nivChargerCorps(fichNiv, nivAct);
    
    fclose(fichNiv);
    return retour;
}


/**********************************************/
int nivChargerCorps(FILE *fNiv, Niveau *nivAct)
{
    int iFich = 0,
        i,
        j,
        nbLignes = 0,
        longLigne = 0,
        maxLongLigne = 0;
    long posFich;
    
    posFich = ftell(fNiv); /* on y reviendra pour inscrire les donnees dans la matrice */
    
    /* On parcourt le corps pour voir le nombre de lignes et la longueur de la ligne la plus longue */
    while((iFich = fgetc(fNiv)) == ' ' || iFich == '#' || iFich == '\n' ||
                  iFich == '*' || iFich == '$' || iFich == '.' || iFich == '@' || iFich == '+')
    {
        longLigne += 1;
        if(iFich == '\n')
        {
            if(longLigne > 1)
                nbLignes += 1;
            if(longLigne - 1 > maxLongLigne)
                maxLongLigne = longLigne - 1;
            longLigne = 0;
        }
    }
    
    /* Allocations */
    nivAct->mat = NULL;
    if((nivAct->mat = malloc(nbLignes * sizeof(char*))) == NULL)
        return ERR_MEMOIRE;
    for(i = 0; i < nbLignes; i++)
    {
        nivAct->mat[i] = NULL;
        if((nivAct->mat[i] = malloc(maxLongLigne * sizeof(char))) == NULL)
            return ERR_MEMOIRE;
    }
    
    /* enregistrement des donnees */
    fseek(fNiv, posFich, SEEK_SET);
    nivAct->posJ.x = -1;
    nivAct->posJ.y = -1;
    nivAct->largMat = maxLongLigne;
    nivAct->hautMat = nbLignes;
    nivAct->nbTrous = 0;
    nivAct->nbCaisses = 0;
    nivAct->nbCaisseTrou = 0;
    nivAct->nbDepl = 0;
    for(i = 0; i < nbLignes; i++)
    {
        j = 0;
        while((iFich = fgetc(fNiv)) != '\n')
        {
            switch(iFich)
            {
                case ' ':
                    nivAct->mat[i][j] = VIDE;
                    break;
                case '#':
                    nivAct->mat[i][j] = MUR;
                    break;
                case '$':
                    nivAct->mat[i][j] = BLOC;
                    nivAct->nbCaisses += 1;
                    break;
                case '.':
                    nivAct->mat[i][j] = TROU;
                    nivAct->nbTrous += 1;
                    break;
                case '*':
                    nivAct->mat[i][j] = BLOC_TROU;
                    nivAct->nbCaisseTrou += 1;
                    break;
                case '@':
                    nivAct->mat[i][j] = JOUEUR;
                    nivAct->posJ.x = j;
                    nivAct->posJ.y = i;
                    break;
                case '+':
                    nivAct->mat[i][j] = JOUEUR_TROU;
                    nivAct->posJ.x = j;
                    nivAct->posJ.y = i;
                    nivAct->nbTrous += 1;
                    break;
            }
            j += 1;
        }
        if(j < maxLongLigne)
             for(; j < maxLongLigne; j++)
                   nivAct->mat[i][j] = VIDE;
        
    }
    /* petites verifs. de bases pour voir si le fichier est correct. */
    if(nivAct->posJ.x < 0 || nivAct->nbCaisses != nivAct->nbTrous || nivAct->nbCaisses == 0)
        return ERR_FICH_INVALIDE;
    else
        return SUCCES;
} 

/**********************************************/
int nivCopier(Niveau *nSrc, Niveau *nDest)
{
    int i;
    
    *nDest = *nSrc; /* d'un coup toutes les petites variables, reste la matrice */
    nDest->mat = NULL;
    if((nDest->mat = malloc(nDest->hautMat * sizeof(char*))) == NULL)
        return ERR_MEMOIRE;
    for(i = 0; i < nDest->hautMat; i++)
    {
        nDest->mat[i] = NULL;
        if((nDest->mat[i] = malloc(nDest->largMat * sizeof(char))) == NULL)
            return ERR_MEMOIRE;
        memcpy(nDest->mat[i], nSrc->mat[i], nDest->largMat * sizeof(char));
    }
    
    return SUCCES;
}

void nivLiberer(Niveau *niv)
{
    int i;
    
    for(i = 0; i < niv->hautMat; i++)
    {
        free(niv->mat[i]);
        niv->mat[i] = NULL;
    }
    free(niv->mat);
}

/**********************************************/
int recupEvent(void)
{
    int ok = 0,
        retour = 0;
    int entree;
    
    while(ok == 0)
    {
        entree = getch();
        switch(entree)
        {
            case KEY_UP:
                ok = 1;
                retour = HAUT;
                break;
            case KEY_LEFT:
                ok = 1;
                retour = GAUCHE;
                break;
            case KEY_RIGHT:
                ok = 1;
                retour = DROITE;
                break;
            case KEY_DOWN:
                ok = 1;
                retour = BAS;
                break;
            case 'q':
                ok = 1;
                retour = QUITTER;
                break;
            case 'r':
                ok = 1;
                retour = RECOMMENCER;
                break;
        }
    }
    return retour;
}

/**********************************************/
void afficherCadre(Niveau nivAct, Pos debutCadre, Pos finCadre)
{
    int i,
        j;
    
    for(i = debutCadre.y; i < finCadre.y; i++)
    {
        move(POS_JEU_Y + i, POS_JEU_X + debutCadre.x);
        for(j = debutCadre.x; j < finCadre.x; j++)
        {
            refresh();
            switch(nivAct.mat[i][j])
            {
                case VIDE:
                    attron(COLOR_PAIR(VIDE_COULEUR));
                    addch(C_VIDE);
                    attroff(COLOR_PAIR(VIDE_COULEUR));
                    break;
                case BLOC:
                    attron(COLOR_PAIR(BLOC));
                    addch(C_BLOC);
                    attroff(COLOR_PAIR(BLOC));
                    break;
                case MUR:
                    attron(COLOR_PAIR(MUR));
                    addch(C_MUR);
                    attroff(COLOR_PAIR(MUR));
                    break;
                case TROU:
                    attron(COLOR_PAIR(TROU));
                    addch(C_TROU);
                    attroff(COLOR_PAIR(TROU));
                    break;
                case JOUEUR:
                    attron(COLOR_PAIR(JOUEUR));
                    addch(C_JOUEUR);
                    attroff(COLOR_PAIR(JOUEUR));
                    break;
                case BLOC_TROU:
                    attron(COLOR_PAIR(BLOC_TROU) | A_BOLD);
                    addch(C_BLOC);
                    attroff(COLOR_PAIR(BLOC_TROU) | A_BOLD);
                    break;
                case JOUEUR_TROU:
                    attron(COLOR_PAIR(JOUEUR_TROU) | A_BOLD);
                    addch(C_JOUEUR);
                    attroff(COLOR_PAIR(JOUEUR_TROU) | A_BOLD);
                    break;
            }
        }
    }
}

/**********************************************/
void afficherTout(Niveau nivAct)
{
    Pos posDep = {0, 0},
        posFin = {nivAct.largMat, nivAct.hautMat};
    
    clear();
    afficherTitre();
    attron(COLOR_PAIR(BLOC) | A_BOLD);
    mvprintw(POS_NOM_Y, POS_NOM_X, nivAct.nom);
    attroff(COLOR_PAIR(BLOC) | A_BOLD);
    attron(COLOR_PAIR(VIDE));
    mvprintw(POS_INFOS_Y, POS_INFOS_X, "Nombre Deplacements : ");
    mvprintw(POS_INFOS_Y + 1, POS_INFOS_X, "Nombre trous restant : ");
    mvprintw(POS_INFOS_Y + 2, POS_INFOS_X, "Nombre caisses : ");
    afficherInfos(nivAct);
    mvprintw(POS_INSTRU_Y, POS_INSTRU_X, "Bouger : fleches");
    mvaddstr(POS_INSTRU_Y + 1, POS_INSTRU_X, "Recommencer : 'r'");
    mvaddstr(POS_INSTRU_Y + 2, POS_INSTRU_X, "Quitter : 'q'");
    attroff(COLOR_PAIR(VIDE));
    afficherCadre(nivAct, posDep, posFin);
    
    refresh();
}

/**********************************************/
void afficherTitre(void)
{
    attron(COLOR_PAIR(TITRE_COULEUR));
    mvaddstr(POS_TITRE_Y, POS_TITRE_X,     "<<<<<<<>>>>>>");
    mvaddstr(POS_TITRE_Y + 1, POS_TITRE_X , "<  SOKOBAN  >");
    mvaddstr(POS_TITRE_Y + 2, POS_TITRE_X, "<<<<<<<>>>>>>");
    attroff(COLOR_PAIR(TITRE_COULEUR));
}

/**********************************************/
void afficherInfos(Niveau nivAct)
{
    mvprintw(POS_INFOS_Y, POS_INFOS_X + 22, "%d   ", nivAct.nbDepl);
    mvprintw(POS_INFOS_Y + 1, POS_INFOS_X + 23, "%d ", nivAct.nbCaisses);
    mvprintw(POS_INFOS_Y + 2, POS_INFOS_X + 17, "%d ", nivAct.nbCaisses + nivAct.nbCaisseTrou);
    refresh();
}

/**********************************************/
void message(char *texte, int nbSec)
{
    clear();
    attron(COLOR_PAIR(BLOC) | A_BOLD);
    mvaddstr(HAUTEUR_FEN / 2, LARGEUR_FEN / 2  - strlen(texte) / 2 - 1, texte);
    attroff(COLOR_PAIR(BLOC) | A_BOLD);
    refresh();
    if(nbSec)
    {
        napms(1000 * nbSec);
        clear();
    }
}

/**********************************************/
